/*
    futebus - FUse TExt BUS - fuse based IPC bus for text lines
    Copyright (C) 2023  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/futebus
    VCS: svn://svn.repo.hu/futebus/trunk
    Contact: see doc/contact.html
*/

#include <errno.h>

#include "futebus.h"
#include "client.h"
#include "channel.h"
#include "message.h"

uint32_t next_channel_id;

static channel_t *path2chan(const char *path)
{
	channel_t *c;

	if (strcmp(path, "_stat") == 0)
		return statch;

	mutex_lock(&channels_lock);
	for(c = gdl_first(&channels); c != NULL; c = gdl_next(&channels, c)) {
		if (strcmp(c->path, path) == 0) {
			mutex_unlock(&channels_lock);
			return c;
		}
	}

	mutex_unlock(&channels_lock);
	return NULL;
}

void chan_destroy_(channel_t *chan)
{
	client_t *client, *cnext;
	message_t *msg, *mnext;

	for(client = gdl_first(&chan->clients); client != NULL; client = cnext) {
		cnext = gdl_next(&chan->clients, client);
		client_free(client);
	}
	for(msg = gdl_first(&chan->messages); msg != NULL; msg = mnext) {
		mnext = gdl_next(&chan->messages, msg);
		message_free(msg);
	}

	free(chan->path);
	mutex_uninit(&chan->lock);

	free(chan);
}

void chan_destroy(channel_t *chan)
{
	/* remove from channel list */
	mutex_lock(&channels_lock);
	if (chan->zombie)
		gdl_remove(&zombie_channels, chan, link);
	else
		gdl_remove(&channels, chan, link);
	mutex_unlock(&channels_lock);

	/* free fields */
	chan_destroy_(chan);
}

int chan_unlink(channel_t *chan)
{
	if (chan->zombie)
		return 0;

	mutex_lock(&channels_lock);
	gdl_remove(&channels, chan, link);
	mutex_unlock(&channels_lock);

	if (chan->clients.first != NULL) {
		mutex_lock(&channels_lock);
		chan->zombie = 1;
		gdl_append(&zombie_channels, chan, link);
		mutex_unlock(&channels_lock);
	}
	else {
		chan_destroy_(chan);
		return 1;
	}

	return 0;
}

/* called after a client closed connection in a channel and the client
   got removed from the channel. Some channels need to be unlinked. This
   function may free(chan). */
static void chan_post_close(channel_t *chan, int client_was_master)
{
	mutex_lock(&chan->lock);
	if (client_was_master && CHANTYPE_NEEDS_MASTER(chan->type)) { /* exists only if it has a master */
		chan_unlink(chan);
		mutex_unlock(&chan->lock);
	}
	else if ((chan->clients.first == NULL) && CHANTYPE_AUTOREMOVE(chan->type)) { /* no more clients */
		chan_destroy(chan);
		/* no need to unlock, chan is destroyed */
	}
}

channel_t *chan_alloc(const char *path)
{
	channel_t *chan = calloc(sizeof(channel_t), 1);
	if (chan == NULL)
		return NULL;
	chan->path = strdup(path);
	chan->type = *path;
	mutex_init(&chan->lock);

	mutex_lock(&channels_lock);
	chan->id = next_channel_id++;
	mutex_unlock(&channels_lock);

	return chan;
}

int chan_open(const char *path, uint64_t *fh, int rd, int wr, int as_master)
{
	channel_t *chan;
	client_t *client;

	switch(*path) {
		case 'U': case 'S': case 'A': case 'B': break;
		default: return -ENXIO; /* invalid type */
	}

	chan = path2chan(path);
	if (chan == NULL) {
		/* if master needs to open the channel first and we are not a master -> error */
		if (!as_master && CHANTYPE_NEEDS_MASTER(*path))
			return -ESPIPE;

		/* allocate new channel */
		chan = chan_alloc(path);
		if (chan == NULL)
			return -ENOMEM;

		mutex_lock(&channels_lock);
		gdl_append(&channels, chan, link);
		mutex_unlock(&channels_lock);
	}
	else if (as_master)
		return -EEXIST; /* a second master attempts to open the channel */

	client = client_alloc_to_chan(chan, rd, wr, as_master);
	if (client == NULL) {
		chan_post_close(chan, 1);
		return -ENOMEM;
	}

	*fh = chan_client_to_fd(chan->id, client->id);
	return 0;
}
